from __future__ import unicode_literals

import datetime
import logging
import re

from mopidy.models import Artist

__all__ = (
    'parse_bitrate',
    'parse_date',
    'parse_length',
    'parse_mtime',
    'parse_track',
    'parse_creator'
)

DURATION_RE = re.compile(r"""
(?:
    (?:(?P<hours>\d+):)?
    (?P<minutes>\d+):
)?
(?P<seconds>\d+)
""", flags=re.VERBOSE)

ISODATE_RE = re.compile(r"""
(\d{4})
(?:\-(\d{2}))?
(?:\-(\d{2}))?
""", flags=re.VERBOSE)

logger = logging.getLogger(__name__)


def parse_bitrate(string, default=None):
    if string:
        try:
            return int(float(string))
        except:
            logger.warn('Invalid Internet Archive bitrate: %r', string)
    return default


def parse_date(string, default=None):
    if string:
        try:
            return '-'.join(ISODATE_RE.match(string).groups('01'))
        except:
            logger.warn('Invalid Internet Archive date: %r', string)
    return default


def parse_length(string, default=None):
    if string:
        try:
            groups = DURATION_RE.match(string).groupdict('0')
            d = datetime.timedelta(**{k: int(v) for k, v in groups.items()})
            return int(d.total_seconds() * 1000)
        except:
            logger.warn('Invalid Internet Archive length: %r', string)
    return default


def parse_mtime(string, default=None):
    if string:
        try:
            return int(string)
        except:
            logger.warn('Invalid Internet Archive mtime: %r', string)
    return default


def parse_track(string, default=None):
    if string:
        try:
            return int(string.partition('/')[0])
        except:
            logger.warn('Invalid Internet Archive track: %r', string)
    return default


def parse_creator(obj, default=None):
    if not obj or obj == 'tmp':
        return default
    elif isinstance(obj, basestring):
        return [Artist(name=obj)]
    else:
        return [Artist(name=name) for name in obj]



from __future__ import unicode_literals

from mopidy import backend

import logging
import uritools

logger = logging.getLogger(__name__)


class InternetArchivePlaybackProvider(backend.PlaybackProvider):

    def change_track(self, track):
        parts = uritools.urisplit(track.uri)
        identifier, filename = parts.getpath(), parts.getfragment()
        t = track.copy(uri=self.backend.client.geturl(identifier, filename))
        return super(InternetArchivePlaybackProvider, self).change_track(t)



from __future__ import unicode_literals

import collections
import logging
import requests
import uritools

BASE_URL = 'http://archive.org/'

logger = logging.getLogger(__name__)


class InternetArchiveClient(object):

    pykka_traversable = True

    def __init__(self, base_url=BASE_URL, retries=0, timeout=None):
        self.base_url = base_url
        self.retries = retries
        self.timeout = timeout
        self.session = requests.Session()

    def search(self, query, fields=None, sort=None, rows=None, start=None):
        response = self._get('/advancedsearch.php', params={
            'q': query,
            'fl[]': fields,
            'sort[]': sort,
            'rows': rows,
            'start': start,
            'output': 'json'
        })
        if response.content:
            return self.SearchResult(response.json())
        else:
            raise self.SearchError(response.url)

    def metadata(self, identifier):
        obj = self._get('/metadata/' + identifier).json()
        if not obj:
            raise LookupError('Internet Archive item %s not found' % identifier)  # noqa
        elif 'error' in obj:
            raise LookupError(obj['error'])
        elif 'result' in obj:
            return obj['result']
        else:
            return obj

    def bookmarks(self, username):
        response = self._get('/bookmarks/' + username, params={
            'output': 'json'
        })
        # requests for non-existant users yield text/xml response
        if response.headers.get('Content-Type') != 'application/json':
            raise LookupError('Internet Archive user %s not found' % username)
        return response.json()

    def geturl(self, identifier, filename=None):
        if filename:
            path = identifier + '/' + uritools.uriencode(filename)
        else:
            path = identifier + '/'
        return uritools.urijoin(self.base_url, '/download/' + path)

    def _get(self, path, params=None):
        url = uritools.urijoin(self.base_url, path)
        retries = self.retries
        timeout = self.timeout

        while True:
            try:
                return self.session.get(url, params=params, timeout=timeout)
            except requests.exceptions.ConnectionError as e:
                if not retries:
                    raise e
                logger.warn('Error connecting to the Internet Archive: %s', e)
                retries -= 1

    class SearchResult(collections.Sequence):

        def __init__(self, result):
            self.query = result['responseHeader']['params']['q']
            self.rowcount = result['response']['numFound']
            self.docs = result['response']['docs']

        def __getitem__(self, key):
            return self.docs[key]

        def __len__(self):
            return len(self.docs)

        def __iter__(self):
            return iter(self.docs)

    class SearchError(Exception):
        pass



from __future__ import unicode_literals

import cachetools
import collections
import itertools
import logging
import operator
import uritools
import re

from mopidy import backend
from mopidy.models import Album, Track, SearchResult, Ref

from . import Extension
from .parsing import *  # noqa
from .query import Query

SCHEME = Extension.ext_name

QUERY_CHAR_RE = re.compile(r'([+!(){}\[\]^"~*?:\\]|\&\&|\|\|)')

QUERY_MAPPING = {
    'any': None,
    'album': 'title',
    'albumartist': 'creator',
    'date': 'date'
}

logger = logging.getLogger(__name__)


def _bookmarks(config):
    if not config['username']:
        return None
    uri = uritools.uricompose(
        scheme=SCHEME,
        userinfo=config['username'],
        host=uritools.urisplit(config['base_url']).host,
        path='/bookmarks/'
    )
    return Ref.directory(name='Archive Bookmarks', uri=uri)


def _cache(cache_size=None, cache_ttl=None, **kwargs):
    """Cache factory"""
    if cache_size is None:
        return None  # mainly for testing/debugging
    elif cache_ttl is None:
        return cachetools.LRUCache(cache_size)
    else:
        return cachetools.TTLCache(cache_size, cache_ttl)


def _ref(metadata):
    identifier = metadata['identifier']
    uri = uritools.uricompose(SCHEME, path=identifier)
    name = metadata.get('title', identifier)
    if metadata.get('mediatype', 'collection') == 'collection':
        return Ref.directory(uri=uri, name=name)
    else:
        return Ref.album(uri=uri, name=name)


def _album(metadata, images=[]):
    identifier = metadata['identifier']
    uri = uritools.uricompose(SCHEME, path=identifier)
    name = metadata.get('title', identifier)
    artists = parse_creator(metadata.get('creator'))
    date = parse_date(metadata.get('date'))
    return Album(uri=uri, name=name, artists=artists, date=date, images=images)


def _track(metadata, file, album):
    identifier = metadata['identifier']
    filename = file['name']
    uri = uritools.uricompose(SCHEME, path=identifier, fragment=filename)
    name = file.get('title', filename)
    return Track(
        uri=uri,
        name=name,
        album=album,
        artists=album.artists,
        track_no=parse_track(file.get('track')),
        date=parse_date(file.get('date'), album.date),
        length=parse_length(file.get('length')),
        bitrate=parse_bitrate(file.get('bitrate')),
        last_modified=parse_mtime(file.get('mtime'))
    )


def _trackkey(track):
    return (track.track_no or 0, track.uri)


def _find(mapping, keys, default=None):
    for key in keys:
        if key in mapping:
            return mapping[key]
    return default


def _quote(term):
    term = QUERY_CHAR_RE.sub(r'\\\1', term)
    # only quote if term contains whitespace, since date:"2014-01-01"
    # will raise an error
    if any(c.isspace() for c in term):
        term = '"' + term + '"'
    return term


def _query(*args, **kwargs):
    terms = []
    for field, value in itertools.chain(args, kwargs.items()):
        if not value:
            continue
        elif isinstance(value, basestring):
            values = [_quote(value)]
        else:
            values = [_quote(v) for v in value]
        if len(values) == 1:
            term = values[0]
        else:
            term = '(%s)' % ' OR '.join(values)
        terms.append(field + ':' + term if field else term)
    return ' AND '.join(terms)


class InternetArchiveLibraryProvider(backend.LibraryProvider):

    root_directory = Ref.directory(
        uri=uritools.uricompose(SCHEME),
        name='Internet Archive'
    )

    def __init__(self, config, backend):
        super(InternetArchiveLibraryProvider, self).__init__(backend)
        self._config = ext_config = config[Extension.ext_name]
        self._search_filter = {
            'format': ext_config['audio_formats'],
            '-collection': ext_config['exclude_collections'],
            '-mediatype': ext_config['exclude_mediatypes']
        }
        self._bookmarks = _bookmarks(ext_config)
        self._cache = _cache(**ext_config)
        self._tracks = {}  # track cache for faster lookup

    def lookup(self, uri):
        try:
            return [self._tracks[uri]]
        except KeyError:
            logger.debug("track lookup cache miss %r", uri)
        try:
            _, _, identifier, _, filename = uritools.urisplit(uri)
            tracks = self._lookup(identifier)
            self._tracks = trackmap = {t.uri: t for t in tracks}
            return [trackmap[uri]] if filename else tracks
        except Exception as e:
            logger.error('Lookup failed for %s: %s', uri, e)
            return []

    def browse(self, uri):
        logger.info('browse %r', uri)
        try:
            parts = uritools.urisplit(uri)
            if not parts.path:
                return self._browse_root()
            elif parts.userinfo:
                return self._browse_bookmarks(parts.userinfo)
            elif parts.path in self._config['collections']:
                return self._browse_collection(parts.path)
            else:
                return self._browse_item(parts.path)
        except Exception as e:
            logger.error('Error browsing %s: %s', uri, e)
            return []

    def search(self, query=None, uris=None):
        try:
            terms = []
            for field, value in (query.iteritems() if query else []):
                if field not in QUERY_MAPPING:
                    return None  # no result if unmapped field
                elif isinstance(value, basestring):
                    terms.append((QUERY_MAPPING[field], value))
                else:
                    terms.extend((QUERY_MAPPING[field], v) for v in value)
            if uris:
                urisplit = uritools.urisplit
                ids = filter(None, (urisplit(uri).path for uri in uris))
                terms.append(('collection', tuple(ids)))
            return self._search(*terms)
        except Exception as e:
            logger.error('Error searching the Internet Archive: %s', e)
            return None

    def find_exact(self, query=None, uris=None):
        try:
            terms = []
            for field, value in (query.iteritems() if query else []):
                if field not in QUERY_MAPPING:
                    return None  # no result if unmapped field
                elif isinstance(value, basestring):
                    terms.append((QUERY_MAPPING[field], value))
                else:
                    terms.extend((QUERY_MAPPING[field], v) for v in value)
            if uris:
                urisplit = uritools.urisplit
                ids = filter(None, (urisplit(uri).path for uri in uris))
                terms.append(('collection', tuple(ids)))
            result = self._search(*terms)
            albums = filter(Query(query, True).match_album, result.albums)
            return result.copy(albums=albums)
        except Exception as e:
            logger.error('Error searching the Internet Archive: %s', e)
            return None

    def refresh(self, uri=None):
        logger.info('Clearing Internet Archive cache')
        self._cache.clear()
        self._tracks.clear()

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_root(self):
        logger.info('browse root')
        collections = self._config['collections']
        refs = [self._bookmarks] if self._bookmarks else []
        docs = {doc['identifier']: doc for doc in self.backend.client.search(
            _query(identifier=collections, mediatype='collection'),
            fields=['identifier', 'title', 'mediatype'],
            rows=len(collections)
        )}
        # return in same order as listed in config
        for id in collections:
            if id in docs:
                refs.append(_ref(docs[id]))
            else:
                logger.warn('Internet Archive collection "%s" not found', id)
        return refs

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_bookmarks(self, username):
        return map(_ref, self.backend.client.bookmarks(username))

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_collection(self, identifier):
        return map(_ref, self.backend.client.search(
            _query(collection=identifier, **self._search_filter),
            fields=['identifier', 'title', 'mediatype'],
            sort=self._config['browse_order'],
            rows=self._config['browse_limit']
        ))

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_item(self, identifier):
        tracks = self._lookup(identifier)
        self._tracks = {t.uri: t for t in tracks}  # cache tracks
        return [Ref.track(uri=t.uri, name=t.name) for t in tracks]

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _lookup(self, identifier):
        client = self.backend.client
        item = client.metadata(identifier)
        files = collections.defaultdict(dict)
        # HACK: not all files have "mtime", but reverse-sorting on
        # filename tends to preserve "l(e)ast derived" versions,
        # e.g. "filename.mp3" over "filename_vbr.mp3"
        key = operator.itemgetter('name')
        for file in sorted(item['files'], key=key, reverse=True):
            original = str(file.get('original', file['name']))
            files[file['format']][original] = file
        images = []
        for file in _find(files, self._config['image_formats'], {}).values():
            images.append(client.geturl(identifier, file['name']))
        album = _album(item['metadata'], images)
        tracks = []
        for file in _find(files, self._config['audio_formats'], {}).values():
            tracks.append(_track(item['metadata'], file, album))
        tracks.sort(key=_trackkey)
        return tracks

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _search(self, *terms):
        result = self.backend.client.search(
            _query(*terms, **self._search_filter),
            fields=['identifier', 'title', 'creator', 'date'],
            sort=self._config['search_order'],
            rows=self._config['search_limit']
        )
        uri = uritools.uricompose(SCHEME, query=result.query)
        return SearchResult(uri=uri, albums=[_album(doc) for doc in result])



from __future__ import unicode_literals

import logging
import pykka

from mopidy import backend

from . import Extension
from .client import InternetArchiveClient
from .library import InternetArchiveLibraryProvider
from .playback import InternetArchivePlaybackProvider

logger = logging.getLogger(__name__)


class InternetArchiveBackend(pykka.ThreadingActor, backend.Backend):

    uri_schemes = [Extension.ext_name]

    def __init__(self, config, audio):
        super(InternetArchiveBackend, self).__init__()
        self.client = InternetArchiveClient(
            config[Extension.ext_name]['base_url'],
            retries=config[Extension.ext_name]['retries'],
            timeout=config[Extension.ext_name]['timeout']
        )
        self.library = InternetArchiveLibraryProvider(config, self)
        self.playback = InternetArchivePlaybackProvider(audio, self)



from __future__ import unicode_literals

import unittest
import pykka

from mopidy_internetarchive.backend import InternetArchiveBackend
from mopidy import core

TEST_ARTIST = 'Martin Luther King'
TEST_ALBUM = 'I Have a Dream'
TEST_DATE = '1963-08-28'


class LibraryTest(unittest.TestCase):
    config = {
        'internetarchive': {
            'base_url': 'http://archive.org',
            'username': None,
            'collections': ('etree', 'audio'),
            'audio_formats': ('MP3'),
            'image_formats': ('JPEG'),
            'browse_limit': None,
            'browse_order': None,
            'search_limit': None,
            'search_order': None,
            'exclude_collections': (),
            'exclude_mediatypes': (),
            'cache_size': 1,
            'cache_ttl': None,
            'retries': 0,
            'timeout': None
        }
    }

    def setUp(self):
        self.backend = InternetArchiveBackend.start(
            config=self.config, audio=None).proxy()
        self.core = core.Core(backends=[self.backend])
        self.library = self.core.library

    def tearDown(self):
        pykka.ActorRegistry.stop_all()

    def test_search_artist(self):
        self.library.search(artist=[TEST_ARTIST])
        # TODO: write tests

    def test_search_album(self):
        self.library.search(album=[TEST_ALBUM])
        # TODO: write tests

    def test_search_date(self):
        self.library.search(date=[TEST_DATE])
        # TODO: write tests



from __future__ import unicode_literals

import unittest

from mopidy_internetarchive import Extension


class ExtensionTest(unittest.TestCase):

    def test_get_default_config(self):
        ext = Extension()

        config = ext.get_default_config()
        self.assertIn('[internetarchive]', config)
        self.assertIn('enabled = true', config)

    def test_get_config_schema(self):
        ext = Extension()

        schema = ext.get_config_schema()
        self.assertIn('base_url', schema)
        self.assertIn('collections', schema)
        self.assertIn('audio_formats', schema)
        self.assertIn('image_formats', schema)
        self.assertIn('browse_limit', schema)
        self.assertIn('browse_order', schema)
        self.assertIn('search_limit', schema)
        self.assertIn('search_order', schema)
        self.assertIn('exclude_collections', schema)
        self.assertIn('exclude_mediatypes', schema)
        self.assertIn('cache_size', schema)
        self.assertIn('cache_ttl', schema)
        self.assertIn('retries', schema)
        self.assertIn('timeout', schema)



from __future__ import unicode_literals

import unittest

from mopidy.models import Artist

from mopidy_internetarchive.parsing import *  # noqa


class ParsingTest(unittest.TestCase):

    def test_parse_bitrate(self):
        self.assertEqual(parse_bitrate(None), None)
        self.assertEqual(parse_bitrate(''), None)
        self.assertEqual(parse_bitrate([]), None)
        self.assertEqual(parse_bitrate('tmp'), None)

        self.assertEqual(parse_bitrate(None, 42), 42)
        self.assertEqual(parse_bitrate('', 42), 42)
        self.assertEqual(parse_bitrate([], 42), 42)
        self.assertEqual(parse_bitrate('tmp', 42), 42)

        self.assertEqual(parse_bitrate('0'), 0)
        self.assertEqual(parse_bitrate('1'), 1)
        self.assertEqual(parse_bitrate('42.123'), 42)

    def test_parse_date(self):
        self.assertEqual(parse_date(None), None)
        self.assertEqual(parse_date(''), None)
        self.assertEqual(parse_date([]), None)
        self.assertEqual(parse_date('tmp'), None)

        self.assertEqual(parse_date(None, 42), 42)
        self.assertEqual(parse_date('', 42), 42)
        self.assertEqual(parse_date([], 42), 42)
        self.assertEqual(parse_date('tmp', 42), 42)

        self.assertEqual(parse_date('2014-02-21T06:00:00Z'), '2014-02-21')
        self.assertEqual(parse_date('2014-02-21'), '2014-02-21')
        self.assertEqual(parse_date('2014-02'), '2014-02-01')
        self.assertEqual(parse_date('2014'), '2014-01-01')

    def test_parse_length(self):
        self.assertEqual(parse_length(None), None)
        self.assertEqual(parse_length(''), None)
        self.assertEqual(parse_length([]), None)
        self.assertEqual(parse_length('tmp'), None)

        self.assertEqual(parse_length(None, 42), 42)
        self.assertEqual(parse_length('', 42), 42)
        self.assertEqual(parse_length([], 42), 42)
        self.assertEqual(parse_length('tmp', 42), 42)

        self.assertEqual(parse_length('0'), 0)
        self.assertEqual(parse_length('1'), 1 * 1000)
        self.assertEqual(parse_length('60'), 60 * 1000)
        self.assertEqual(parse_length('1:00'), 60 * 1000)
        self.assertEqual(parse_length('60:00'), 3600 * 1000)
        self.assertEqual(parse_length('1:00:00'), 3600 * 1000)

    def test_parse_mtime(self):
        self.assertEqual(parse_mtime(None), None)
        self.assertEqual(parse_mtime(''), None)
        self.assertEqual(parse_mtime([]), None)
        self.assertEqual(parse_mtime('tmp'), None)

        self.assertEqual(parse_mtime(None, 42), 42)
        self.assertEqual(parse_mtime('', 42), 42)
        self.assertEqual(parse_mtime([], 42), 42)
        self.assertEqual(parse_mtime('tmp', 42), 42)

        self.assertEqual(parse_mtime('0'), 0)
        self.assertEqual(parse_mtime('1'), 1)

    def test_parse_track(self):
        self.assertEqual(parse_track(None), None)
        self.assertEqual(parse_track(''), None)
        self.assertEqual(parse_track([]), None)
        self.assertEqual(parse_track('tmp'), None)

        self.assertEqual(parse_track(None, 42), 42)
        self.assertEqual(parse_track('', 42), 42)
        self.assertEqual(parse_track([], 42), 42)
        self.assertEqual(parse_track('tmp', 42), 42)

        self.assertEqual(parse_track('0'), 0)
        self.assertEqual(parse_track('1'), 1)
        self.assertEqual(parse_track('1/100'), 1)
        self.assertEqual(parse_track('1/tmp'), 1)

    def test_parse_creator(self):
        artist = Artist(name='foo')

        self.assertEqual(parse_creator(None), None)
        self.assertEqual(parse_creator(''), None)
        self.assertEqual(parse_creator([]), None)
        self.assertEqual(parse_creator('tmp'), None)

        self.assertEqual(parse_creator(None, [artist]), [artist])
        self.assertEqual(parse_creator('', [artist]), [artist])
        self.assertEqual(parse_creator([], [artist]), [artist])
        self.assertEqual(parse_creator('tmp', [artist]), [artist])

        self.assertEqual(parse_creator('foo'), [artist])
        self.assertEqual(parse_creator(['foo']), [artist])
        self.assertEqual(parse_creator(['foo', 'foo']), [artist, artist])
